import * as METADATA from '../constants';

import {
  ClassConstructor,
  DecoratorType,
  PouchScope,
  POUCH_DEFAULT_QUALIFIER,
} from '../constants';

import PouchDefinition from '../config/PouchDefinition';

import PouchInjectDefinition, {
  InjectItem,
} from '../config/PouchInjectDefinition';

import InjectConfig from '../decorator/InjectConfig';

class PouchFactory {
  private pouchDefinitionMap: Map<Function, PouchDefinition>;
  private pouchInstanceMap: Map<Function, Map<string | symbol, any>>;

  public constructor() {
    this.pouchDefinitionMap = new Map();
    this.pouchInstanceMap = new Map();
  }

  /**
   * getPouch
   */
  public getPouch<T>(type: ClassConstructor<T>, ...args: any[]): T {
    let qualifier =  POUCH_DEFAULT_QUALIFIER;
    return this.getPouchByQualifier(type, qualifier, args);
  }

  /**
   * getPouchByQualifier
   */
  public getPouchByQualifier<T>(type: ClassConstructor<T>, qualifier: string | symbol, ...args: any[]): T {
    let clazz: Function = type.prototype;

    let instance = this._getPouch(clazz, qualifier, args);

    return instance;
  }


  /**
   * registerPouch
   */
  public registerPouch(pouch: any, qualifier: string | symbol = POUCH_DEFAULT_QUALIFIER): void {
    let clazz = pouch.prototype;
    this._putPouchInstance(clazz, qualifier, pouch);
  }

  /**
   * registerPouch
   */
  public registerPouchDefinition<T>(type: ClassConstructor<T>, pouchDefinition: PouchDefinition): void {
    let clazz: Function = type.prototype;
    if (this.pouchDefinitionMap.has(clazz)) {
      throw new Error(`Pouch Definition of [${clazz.name}] is already exist.`);
    }
    this.pouchDefinitionMap.set(clazz, pouchDefinition);
  }

  protected getPouchDefinition(clazz: Function): PouchDefinition {
    let pouchDefinition: PouchDefinition = this.pouchDefinitionMap.get(clazz);
    if (undefined !== pouchDefinition) {
      return pouchDefinition;
    }

    if (Reflect.hasOwnMetadata(METADATA.POUCH_DEFINITION, clazz) !== true) {
      return null;
    }

    pouchDefinition = Reflect.getMetadata(METADATA.POUCH_DEFINITION, clazz);
    this.pouchDefinitionMap.set(clazz, pouchDefinition);

    return pouchDefinition;

  }

  protected getPouchInjectDefinition(clazz: Function): PouchInjectDefinition {
    if (Reflect.hasOwnMetadata(METADATA.POUCH_INJECT_DEFINITION, clazz) !== true) {
      return null;
    }

    let pouchInjectDefinition: PouchInjectDefinition = Reflect.getMetadata(METADATA.POUCH_INJECT_DEFINITION, clazz);

    return pouchInjectDefinition;
  }

  private _getPouch(clazz: Function, qualifier: string | symbol, ...args: any[]): any {
    let pouchDefinition: PouchDefinition = this.getPouchDefinition(clazz);

    let instance: any;
    if (null == pouchDefinition) {
      if (this._hasPouchInstance(clazz, qualifier)) {
        instance = this._getPouchInstance(clazz, qualifier);
      } else {
        instance = this._newInstance(clazz, pouchDefinition, args);
      }
    } else {
      if (pouchDefinition.scope === PouchScope.PROTOTYPE) {
        instance = this._newInstance(clazz, pouchDefinition, args);
      } else {
        if (this._hasPouchInstance(clazz, qualifier)) {
          return this._getPouchInstance(clazz, qualifier);
        }
        instance = this._newInstance(clazz, pouchDefinition, args);
      }
    }

    return instance;
  }

  private _newInstance(clazz: Function, pouchDefinition: PouchDefinition, ...args: any[]): any {
    let instance: any;
    instance = Object.create(clazz);
    instance.constructor.apply(instance, args);

    this._injectDependency(instance, clazz);

    return instance;
  }

  private _injectDependency(target: object, clazz: Function): void {
    if (clazz.constructor === Object) {
      return;
    }

    let pouchInjectDefinition: PouchInjectDefinition = this.getPouchInjectDefinition(clazz);
    if (null !== pouchInjectDefinition) {
      let injectors: Map<string | symbol, InjectItem> = pouchInjectDefinition.injectors;
      let propertyDescriptor: any;
      injectors.forEach((injectItem, key, map) => {
        propertyDescriptor = Object.getOwnPropertyDescriptor(target, key);

        switch (injectItem.decoratorType) {
        case DecoratorType.PROPERTY:
          this._injectDependencyProperty(target, clazz, key, propertyDescriptor, injectItem.propertyConfig);
          break;
        case DecoratorType.PARAMETER:
          this._injectDependencyParameter(target, propertyDescriptor, injectItem.parameterConfigMap);
          break;
        default:
          break;
        }
      });
    }
    this._injectDependency(target, Object.getPrototypeOf(clazz));
  }

  private _injectDependencyProperty(target: object, clazz: Function, propertyKey: string | symbol,
                                    propertyDescriptor: PropertyDescriptor, injectConfig: InjectConfig): void {
    let instance = this.getPouch(injectConfig.type, injectConfig.qualifier);
    if (injectConfig.required && undefined === instance) {
      throw new Error(`Pouch which used by [${clazz.name}.${propertyKey}] is not exist in the context.`);
    }
    target[propertyKey] = instance;
  }

  private _injectDependencyParameter(target: object, propertyDescriptor: PropertyDescriptor,
                                     parameterConfigs: Map<number, InjectConfig>): void {
    console.log('');
  }

  private _hasPouchInstance(clazz: Function, qualifier: string | symbol): boolean {
    if (!this.pouchInstanceMap.has(clazz) || !this.pouchInstanceMap.get(clazz).has(qualifier)) {
      return false;
    }
    return true;
  }

  private _getPouchInstance(clazz: Function, qualifier: string | symbol): any {
    if (!this._hasPouchInstance(clazz, qualifier)) {
      return null;
    }
    return this.pouchInstanceMap.get(clazz).get(qualifier);
  }

  private _putPouchInstance(clazz: Function, qualifier: string | symbol, instance: any): void {
    let instanceMap: Map<string | symbol, any> = this.pouchInstanceMap.get(clazz);

    if (undefined === instanceMap) {
      instanceMap = new Map();
      this.pouchInstanceMap.set(clazz, instanceMap);
    }

    instanceMap.set(qualifier, instance);
  }

}

export default PouchFactory;
